/*
 * FAPG 0.0 released under GPL
 * http://royale.zerezo.com/fapg/
 */

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <dirent.h>
#include <sys/types.h>
#include <string.h>
#include <limits.h>

#define MAX 10240

int debug=0;
int format=0; /* 0 = m3u ; 1 = pls ; 2 = html */
unsigned char *prefix="";
int recursive=0;
int separator='/';
int skip=0;
int windows=0;
unsigned char buffer[MAX];

int counter=0;

unsigned char artist[1024];
unsigned char title[1024];
int time;

unsigned char unix2dos[256]={0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,70,35,36,37,38,39,40,41,82,43,44,45,46,47,48,49,50,51,52,53,54,55,56,57,84,59,36,61,65,71,64,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,91,36,93,94,95,96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,112,113,114,115,116,117,118,119,120,121,122,123,36,125,126,127,199,252,233,226,228,224,229,231,234,235,232,239,238,236,196,197,201,230,198,244,246,242,251,249,255,214,220,248,163,216,215,131,225,237,243,250,241,209,170,186,191,174,172,189,188,161,171,187,166,166,166,166,166,193,194,192,169,166,166,43,43,162,165,43,43,45,45,43,45,43,227,195,43,43,45,45,166,45,43,164,240,208,202,203,200,105,205,206,207,43,43,166,220,166,204,175,211,223,212,210,245,213,181,254,222,218,219,217,253,221,175,180,173,177,61,190,182,167,247,184,176,168,183,185,179,178,166,160};

void usage()
{
  fprintf(stderr,"Usage >> fapg [-backslash] [-debug] [-format=m3u|pls|html] [-output=/path/to/file.m3u] [-prefix=/the/prefix] [-recursive] [-windows] /path/to/mp3/dir1 [/path/to/mp3/dir2 ...]\n");
  exit(1);
}

void parse_options(int argc,char **argv)
{
  static struct option long_options[]=
  {
    {"backslash",no_argument,&separator,'\\'},
    {"debug",no_argument,&debug,1},
    {"format",required_argument,0,'f'},
    {"output",required_argument,0,'o'},
    {"prefix",required_argument,0,'p'},
    {"recursive",no_argument,&recursive,1},
    {"windows",no_argument,&windows,1}
  };
  int c;
  int option_index=0;
  while ((c=getopt_long_only(argc,argv,"",long_options,&option_index))!=-1)
  {
    switch(c)
    {
      case 0:
        break;
      case 'f':
        if (strcmp(optarg,"m3u")==0)  format=0; else
        if (strcmp(optarg,"pls")==0)  format=1; else
        if (strcmp(optarg,"html")==0) format=2; else
        usage();
        break;
      case 'o':
        close(1);
        if (fopen(optarg,"w")==NULL) { fprintf(stderr,"Error >> unable to open output file : %s\n",optarg); exit(2); }
        break;
      case 'p':
        prefix=malloc(strlen(optarg)+1);
        strcpy(prefix,optarg);
        break;
      default:
        usage();
    }
  }
}

void parse_mp3(unsigned char *file)
{
  int bitrates[2][3][15]=
    {{{0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448},
      {0, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384},
      {0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320}},
     {{0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256},
      {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160},
      {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160}}};
  FILE *fic;
  unsigned char *c;
  int lus;

  if (debug) fprintf(stderr,"Debug >> parsing mp3 : %s\n",file);

  /* read header */
  if ((fic=fopen(file,"r"))==NULL) { fprintf(stderr,"Warning >> can't open file : %s\n",file); return; }
  lus=fread(buffer,1,MAX,fic);
  c=buffer;  

  /* try ID3v2 */
  if (buffer[0]=='I' && buffer[1]=='D' && buffer[2]=='3')
  {
    int size;
    int version;
    version=*(buffer+3);
    if (version<2 || version>4)
      fprintf(stderr,"Warning >> ID3 v2.%d not implemented ! trying anyway : %s\n",version,file);
    if (*(buffer+5)!=0)
      fprintf(stderr,"Warning >> specials headers not implemented (%d) ! trying anyway : %s\n",*(buffer+5),file);
    c=buffer+6;
    size=(*c<<24)+(*(c+1)<<16)+(*(c+2)<<8)+(*(c+3));
    if (size>lus) size=lus;
    c+=4;
    if (version==2) while (c<buffer+size-5)
    {
      int size=(*(c+3)<<16)+(*(c+4)<<8)+(*(c+5));
      if (*c==0) break;
      if (strncmp(c,"TT2",3)==0)
      {
        strncpy(title,c+7,size-1);
        title[size-1]='\0';
      }
      if (strncmp(c,"TP1",3)==0)
      {
        strncpy(artist,c+7,size-1);
        artist[size-1]='\0';
      }
      c+=size+6;
    }
    if (version==3 || version==4) while (c<buffer+size-7)
    {
      int size=(*(c+4)<<24)+(*(c+5)<<16)+(*(c+6)<<8)+(*(c+7));
      if (*c==0) break;
      if (strncmp(c,"TIT2",4)==0)
      {
        strncpy(title,c+11,size-1);
        title[size-1]='\0';
      }
      if (strncmp(c,"TPE1",4)==0)
      {
        strncpy(artist,c+11,size-1);
        artist[size-1]='\0';
      }
      c+=size+10;
    }
  }
  
  while (c<buffer+lus-10)
  {
    if (*c==0xFF && (*(c+1)&0xF0)==0xF0)
    {
      int version;
      int lay;
      int bitrate_index;
      int bitrate;
      version=2-(*(c+1)>>3&1);
      lay=4-(*(c+1)>>1&3);
      bitrate_index=*(c+2)>>4&0xF;
      if (version>=1 && version<=2 && lay-1>=0 && lay-1<=2 && bitrate_index>=0 && bitrate_index<=14)
        bitrate=bitrates[version-1][lay-1][bitrate_index];
      else
        bitrate=0;
      if (bitrate!=0)
      {
        fseek(fic,0,SEEK_END);
        time=(ftell(fic)+buffer-c)/125/bitrate;
      }
      else
        time=0;
      break;
    }
    c++;
  }

  /* try ID3v1 */
  if (strlen(artist)==0 && strlen(title)==0)
  {
    fseek(fic,-128,SEEK_END);
    lus=fread(buffer,1,128,fic);
    if (lus==128 && buffer[0]=='T' && buffer[1]=='A' && buffer[2]=='G')
    {
      strncpy(title,buffer+3,30);
      title[30]='\0';
      c=title+29;
      while (c>title && *c==' ') *(c--)='\0';
      strncpy(artist,buffer+33,30);
      artist[30]='\0';
      c=artist+29;
      while (c>artist && *c==' ') *(c--)='\0';
    }
  }

  fclose(fic);
}  

void parse_ogg(unsigned char *file)
{
  FILE *fic;
  unsigned char *c;
  int lus;
  int sample_rate;
  int samples;

  if (debug) fprintf(stderr,"Debug >> parsing ogg : %s\n",file);

  /* read header */
  if ((fic=fopen(file,"r"))==NULL) { fprintf(stderr,"Warning >> can't open file : %s\n",file); return; }
  lus=fread(buffer,1,MAX,fic);

  /* try Ogg */
  if (buffer[0]!='O' && buffer[1]!='g' && buffer[2]!='g')
  {
    fprintf(stderr,"Warning >> not a Ogg header : %s\n",file);
    return;
  }
  
  c=buffer+0x28;
  sample_rate=(*c)+(*(c+1)<<8)+(*(c+2)<<16)+(*(c+3)<<24);

  while (c<buffer+lus-10)
  {
    int size;
    if (strncmp(c,"TITLE=",6)==0)
    {
      size=*(c-4)+(*(c-3)<<8)+(*(c-2)<<16)+(*(c-1)<<24);
      strncpy(title,c+6,size-6);
      title[size-6]='\0';
      c+=size;
    }
    if (strncmp(c,"ARTIST=",7)==0)
    {
      size=*(c-4)+(*(c-3)<<8)+(*(c-2)<<16)+(*(c-1)<<24);
      strncpy(artist,c+7,size-7);
      artist[size-7]='\0';
      c+=size;
    }
    c++;
  }
  
  fseek(fic,-MAX,SEEK_END);
  lus=fread(buffer,1,MAX,fic);
  c=buffer+lus-1;
  while (strncmp(c,"OggS",4)!=0 && c>buffer) c--;
  if (c!=buffer)
  {
    c+=6;
    samples=(*c)+(*(c+1)<<8)+(*(c+2)<<16)+(*(c+3)<<24);
    time=samples/sample_rate;
  }
  
  fclose(fic);
}

void parse_directory(unsigned char *path)
{
  DIR *dirp;
  struct dirent *dp;
  unsigned char newpath[PATH_MAX];
  unsigned char *c;

  void print_faketitle()
  {
    c=newpath+strlen(newpath);
    while (c>newpath && *c!='/') c--;
    while (c<newpath+strlen(newpath)-5)
    {
      c++;
      if (*c=='_') putchar(' '); else if (windows) putchar(unix2dos[*c]); else putchar(*c);
    }
  }
  
  void print_path()
  {
    printf(prefix);
    for (c=newpath+skip;*c!='\0';c++) if (*c=='/') putchar(separator); else if (windows) putchar(unix2dos[*c]); else putchar(*c);
  }

  if (debug) fprintf(stderr,"Debug >> parsing directory : %s\n",path);
  if ((dirp=opendir(path))==NULL) { fprintf(stderr,"Warning >> can't open directory : %s\n",path); return; }
  while ((dp=readdir(dirp))!=NULL)
  {
    sprintf(newpath,"%s/%s",path,dp->d_name);
    if (recursive && dp->d_type & DT_DIR && strcmp(dp->d_name,".")!=0 && strcmp(dp->d_name,"..")!=0)
    {
      parse_directory(newpath);
    }
    if (dp->d_type & DT_REG)
    {
      unsigned char ext[5];
      int i;
      for (i=0;i<5;i++) ext[i]=tolower(dp->d_name[strlen(dp->d_name)-4+i]);
      artist[0]='\0';
      title[0]='\0';
      time=-2;
      if (strcmp(".mp3",ext)==0) { time=-1; parse_mp3(newpath); }
      if (strcmp(".ogg",ext)==0) { time=-1; parse_ogg(newpath); }
      if (strcmp(".wav",ext)==0) { time=-1; /* parse_wav(newpath); */ }
      
      if (time!=-2) /* is it an audio file ? */
      {
        counter++;
        switch (format)
        {
          case 0:
            if (time!=-1)
            {
              printf("#EXTINF:%d,",time);
              if (strlen(artist)==0 && strlen(title)==0) print_faketitle();
              else printf("%s - %s",artist,title);
              putchar('\n');
            }
            print_path();
            putchar('\n');
            break;
          case 1:
            printf("File%d=",counter);
            print_path();
            putchar('\n');
            printf("Title%d=",counter);
            if (strlen(artist)==0 && strlen(title)==0) print_faketitle();
            else printf("%s - %s",artist,title);
            putchar('\n');
            if (time!=-1) printf("Length%d=%d\n",counter,time);
            break;
          case 2:
            printf("<tr><td>%d</td><td>%s</td><td>%s</td><td>",counter,artist,title);
            if (time==-1) printf("?"); else printf("%d",time);
            printf("</td></tr>\n");
            break;
        }
      }
    }
  }
  closedir(dirp);
}

int main(int argc,char **argv)
{
  parse_options(argc,argv);
  if (optind==argc) usage();
  switch (format)
  {
    case 0:
      printf("#EXTM3U\n");
      break;
    case 1:
      printf("[playlist]\n");
      break;
    case 2:
      printf("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\n\n<html>\n\n<head>\n<title>Playlist generated by FAPG 0.0</title>\n<meta http-equiv=\"Content-Type\" content=\"text/html; charset=iso-8859-1\" />\n<link rel=\"stylesheet\" type=\"text/css\" href=\"fapg.css\" />\n</head>\n\n<body>\n\n<h1>Playlist</h1>\n\n<table>\n<tr><th>Entry</th><th>Artist</th><th>Title</th><th>Length</th></tr>\n");
      break;
  }
  for (;optind<argc;optind++)
  {
    if (strlen(prefix)!=0) skip=strlen(argv[optind]);
    parse_directory(argv[optind]);
  }
  switch(format)
  {
    case 1:
      printf("NumberOfEntries=%d\nVersion=2\n",counter);
      break;
    case 2:
      printf("</table>\n\n<p>Playlist generated by <a href=\"http://royale.zerezo.com/fapg/\">FAPG 0.0</a></p>\n\n</body>\n\n</html>");
      break;
  }
  exit(0);
}
